(**********************************************
	COMS W4115 PLT - 2014 Spring Term Project
	HL-HDL compiler: translate.ml
		Name:	Woon Lee
		UNI:	wgl2107
***********************************************)

open Ast

module StringMap = Map.Make(String)

(* Symbol table: Information about all the names in scope *)
type env = {
	function_decl	: string StringMap.t;	(* function signature *)
	global_var		: v_type StringMap.t;	(* global variables name and types *)
	local_var		: v_type StringMap.t;	(* local & function arg variable names and types *)
}

(* output file header *)
let file_header = "----------------------------------------------------\n" ^
				  "--  COMS W4115 PLT - 2014 Spring Term Project\n" ^
				  "--              HL-HDL compiler\n" ^
				  "--             Woon Lee (wgl2107)\n" ^
				  "----------------------------------------------------\n\n"
(* VHDL header *)
let vhd_header_common = "library IEEE;\n" ^ "use IEEE.STD_LOGIC_1164.all;\n" ^
						"use IEEE.STD_LOGIC_ARITH.ALL;\n" ^ "use IEEE.STD_LOGIC_UNSIGNED.ALL;\n"
let vhd_header_opt = "use work.globals.all;\n" ^ "use work.lib_comp.all;"

(* Port name lists for built-in functions *)
let lat_port = ["RST => "; "EN => "; "CLK => ";"LAT_IN => ";"LAT_OUT => "]
let mux_port = ["MUX_OUT => "; "EN => "; "OUT_SEL => ";"MUX_IN0 => ";"MUX_IN1 => ";"MUX_IN2 => ";"MUX_IN3 => "]

(* extract formal info from function declaration *)
let string_of_formal formal =
	"\t" ^ formal.formal_name ^ " : " ^ Ast.string_of_v_type formal.formal_type ^ " " ^
	match (fst formal.formal_range) with 
		-1	-> " std_logic"
		| _	-> let first = fst formal.formal_range and last = snd formal.formal_range in
				if (first >= last) then "std_logic_vector(" ^ string_of_int first ^ " downto " ^ string_of_int last ^ ")"
										else "std_logic_vector(" ^ string_of_int last ^ " downto " ^ string_of_int first ^ ")"
										
(* extract formal info from function declaration -  for test bench frame generation *)
let string_of_formal_tb formal =
	"\tsignal " ^ formal.formal_name ^ " : " ^
	match (fst formal.formal_range) with 
		-1	-> " std_logic"
		| _	-> let first = fst formal.formal_range and last = snd formal.formal_range in
				if (first >= last) then "std_logic_vector(" ^ string_of_int first ^ " downto " ^ string_of_int last ^ ")"
										else "std_logic_vector(" ^ string_of_int last ^ " downto " ^ string_of_int first ^ ")"

(* extract the name and range of variable *)
let string_of_variable variable =
	"\t" ^ Ast.string_of_v_type variable.var_type ^ " " ^ variable.var_name ^ " : " ^ 
	(match (fst variable.var_range) with 
		-1	-> " std_logic"
		| _	-> let first = fst variable.var_range and last = snd variable.var_range in
				if (first >= last) then "std_logic_vector(" ^ string_of_int first ^ " downto " ^ string_of_int last ^ ")"
										else "std_logic_vector(" ^ string_of_int last ^ " downto " ^ string_of_int first ^ ")" )^
	(match variable.var_value with
		"none"	-> ";\n"
		| _		-> let value_type = String.sub variable.var_value 1 1 in
					match value_type with
					"x"	  -> " := x\"" ^ String.sub variable.var_value 2 ((String.length variable.var_value)-2) ^ "\";\n"
					| "b" -> " := \"" ^ String.sub variable.var_value 2 ((String.length variable.var_value)-2) ^ "\";\n"
					| _   -> " := " ^ variable.var_value ^ ";\n"	)
										
(* Extract variable type and name *)
let rec enum_vdecl = function
	  []		-> []
	| hd::tl	-> (hd.var_type, hd.var_name)::enum_vdecl tl
	
(* Extract formal type and name *)
let rec enum_fdecl = function
	  []		-> []
	| hd::tl	-> (hd.formal_type, hd.formal_name)::enum_fdecl tl

(* Extract function type and name *)
let rec enum_func = function
	  []		-> []
	| hd::tl	-> (Ast.string_of_funtype hd.func_type, hd.func_name)::enum_func tl

(* Extract function name with formals *)
let rec enum_f_formal = function
	  []		-> []
	| hd::tl	-> (hd.formals, hd.func_name)::enum_f_formal tl

(* Extract variable name and width *)
let rec var_width = function
	  []		-> []
	| hd::tl	-> let width = ((fun x -> if x < 0 then -x else x) ((fst hd.var_range)-(snd hd.var_range)))+1 in
					((string_of_int width, hd.var_name))::var_width tl

(* Extract formal name and width *)
let rec formal_width = function
	  []		-> []
	| hd::tl	-> let width = ((fun x -> if x < 0 then -x else x) ((fst hd.formal_range)-(snd hd.formal_range)))+1 in
					((string_of_int width, hd.formal_name))::formal_width tl

(* val string_map_pairs StringMap 'a -> (int * 'a) list -> StringMap 'a *)
let string_map_pairs map pairs =
	List.fold_left (fun m (i, n) -> StringMap.add n i m) map pairs
	
(* Translate the AST to vhdl code *)
let trans_to_vhdl (globals, functions) out_opt = 

	(* Allocate "addresses" for each global variable *)
	(* (variable type, variable name) pair for global variables *)
	let global_variables = string_map_pairs StringMap.empty (enum_vdecl globals) in
	(* (variable width, variable name) pair for global variables *)
	let global_var_widths = string_map_pairs StringMap.empty (var_width globals) in
		(* (function type, function name) pairs *)
		let function_decls = string_map_pairs StringMap.empty (enum_func functions) in
			(* (function formals, function name) pairs *)
			let function_formals = string_map_pairs StringMap.empty (enum_f_formal functions) in
		
	(* Translate a function in AST to a list of bytecode statements *)
	let translate env func_decl = 
		(* (variable type, variable name) pair for local variables *)
		let local_variables = enum_vdecl func_decl.locals
		(* (variable width, variable name) pair for local variables *)
		and local_var_widths = var_width func_decl.locals
		(* (variable type, variable name) pair for formals *)
		and formal_variables = enum_fdecl func_decl.formals
		(* (variable width, variable name) pair for formals *)
		and formal_var_widths = formal_width func_decl.formals in
		
    let env = { env with local_var = string_map_pairs StringMap.empty (local_variables @ formal_variables) }
		and variable_widths = string_map_pairs StringMap.empty (local_var_widths @ formal_var_widths) in
	    
	(* Extract variable type for type checking *)
	let expr_type e =
		match e with
			  Id(id)		-> if (StringMap.mem id global_variables) then
									string_of_v_type (StringMap.find id global_variables)
								else if (StringMap.mem id env.local_var) then
									string_of_v_type (StringMap.find id env.local_var)
										else ""
			| Single (id, index)	-> if (StringMap.mem id global_variables) then
											string_of_v_type (StringMap.find id global_variables)
										else if (StringMap.mem id env.local_var) then
											string_of_v_type (StringMap.find id env.local_var)
											else ""
			| Range (id, first, last)	-> if (StringMap.mem id global_variables) then
												string_of_v_type (StringMap.find id global_variables)
											else if (StringMap.mem id env.local_var) then
												string_of_v_type (StringMap.find id env.local_var)
												else ""
			| _		-> ""
	
	in
	(* evaluate an expression for type checking *)
	let rec eval_expr e =
		match e with
			  Lit_int(l)	-> 0
			| Lit_hex(l)	-> (String.length l - 2)*4
			| Lit_bin(l)	-> (String.length l - 2)
			| Id(id)		-> if (StringMap.mem id global_var_widths)
								then int_of_string (StringMap.find id global_var_widths)	
								else if (StringMap.mem id variable_widths)
									then int_of_string (StringMap.find id variable_widths)
									else  -1								
			| Binop (e1, oper, e2)	-> let e1_length = eval_expr e1 and e2_length = eval_expr e2 in
										if (e1_length == e2_length) then e1_length
										else 
											(if (e1_length == 0) then e2_length
											else (if (e2_length == 0) then e1_length
											else raise (Failure ("type mismatch " ^ Ast.string_of_expr e1 ^ ":" ^ Ast.string_of_expr e2))))
			| Single (id, index)	-> 1
			| Range (id, first, last)	-> if (first >= last) then first-last+1	else last-first+1
			| Braces (e)	-> eval_expr e
			| Assign (e1, e2)	-> let eval_e1 = eval_expr e1 and eval_e2 = eval_expr e2 in
									if (eval_e2 == 0) then eval_e1 else 
									(if (eval_e1 == eval_e2) then eval_e1
										else -1)
			| Call (inst_id, f_id, actuals)	-> 0
			| Bcall (inst_id, btif_id, actuals)	-> 
					let btif_name = Ast.string_of_btif btif_id in
					match btif_name with
					  "Inv"	-> eval_expr (List.hd actuals)
					| "Buf"	-> eval_expr (List.hd actuals)
					| "Lat"	-> 0
					| "Mux"	-> 0
					| _	-> raise (Failure ("undefined built-in function"))
	
	in
	(* Translate an expression *)
	let rec expr e =
		match e with
			  Lit_int(l)	-> string_of_int l
			| Lit_hex(l)	-> let bh_val = String.sub l 2 (String.length l - 2) in
								"x" ^ "\"" ^ bh_val ^ "\""
			| Lit_bin(l)	-> let bh_val = String.sub l 2 ((String.length l) - 2) in
								if (String.length bh_val == 1) then "'" ^ bh_val ^ "'" else "\"" ^ bh_val ^ "\""
			| Id(id)		-> if ((StringMap.mem id env.local_var) || (StringMap.mem id env.global_var))
							   then id else raise (Failure ("undeclared variable: " ^ id))
			| Binop (e1, oper, e2)	->
									expr e1 ^	(match oper with
												  And	-> " AND "
												| Or	-> " OR "
												| Equal	-> " = "
												| Nequal-> " /= ") ^ (expr e2)
			| Single (id, index)	-> id ^ "(" ^ string_of_int index ^ ")"
			| Range (id, first, last)	-> id ^ (if (first >= last) then "(" ^ string_of_int first ^ " downto " ^ string_of_int last ^ ")"
										else "(" ^ string_of_int last ^ " downto " ^ string_of_int first ^ ")")
			| Braces (e)	-> "(" ^ expr e ^ ")"
			| Assign (e1, e2)	-> (match expr_type e1 with 
									"constant" -> raise (Failure ("illegal assignment to a constant: " ^ 
														Ast.string_of_expr e1 ^ " = " ^ Ast.string_of_expr e2))
									| _ -> expr e1 ^ " <= " ^ expr e2)
			| Call (inst_id, f_id, actuals)	-> if not (StringMap.mem f_id env.function_decl) then 
											raise (Failure ("undefined function: " ^ f_id))	
										else
										  let f_formal = (StringMap.find f_id function_formals) in 
											if ((List.length f_formal) != (List.length actuals)) 
											then raise (Failure ("number of fucntion call argument mismatch: " ^ f_id))
											else (
													match inst_id with
													"none"	-> raise (Failure ("Function instantiation id required: " ^ f_id))
													| _		-> inst_id^"_"^f_id^" : "^f_id^" port map (\n\t\t" ^ 
														(String.concat ",\n\t\t" (List.map2 (fun x y -> x ^ y) 
														(List.map (fun x -> x.formal_name^" => ") f_formal) (List.map expr actuals)))^"\n\t)"
												)
			| Bcall (inst_id, btif_id, actuals)	-> 
					let btif_name = Ast.string_of_btif btif_id in
					(
					match btif_name with
					  "Inv"	-> if ((List.length actuals) != 1) then raise (Failure ("Function takes ONE argument: " ^ btif_name))
								else "NOT " ^ String.concat "" (List.map expr actuals) 
					| "Buf"	-> if ((List.length actuals) != 1) then raise (Failure ("Function takes ONE argument: " ^ btif_name))
								else String.concat "" (List.map expr actuals)
					| "Lat"	-> if ((List.length actuals) != 5) then raise (Failure ("Function takes FIVE arguments: " ^ btif_name))
								else (
										let data_width = eval_expr (List.hd (List.rev actuals)) in
										let fcall_name = (if (data_width > 1) then 
															" : "^"LATCH_n_bit\n \tgeneric map(num_bits => " ^
															string_of_int data_width ^ ")\n \tport map (\n\t\t"
														 else " : "^"LATCH port map (\n\t\t") in
										match inst_id with
										"none"	-> raise (Failure ("Function instantiation id required: Lat"))
										| _		-> "\n\t"^inst_id^fcall_name ^
											(String.concat ",\n\t\t" (List.map2 (fun x y -> x^y) lat_port (List.map expr actuals)))^" \n\t)"
									 )
					| "Mux"	-> if ((List.length actuals) != 7) then raise (Failure ("Function takes SEVEN arguments: " ^ btif_name))
								else (
										let data_width = eval_expr (List.hd (List.rev actuals)) in
										let fcall_name = (if (data_width > 1) then 
															" : "^"MUX_n_bit\n \tgeneric map(num_bits => " ^
															string_of_int data_width ^ ")\n \tport map (\n\t\t"
														 else " : "^"MUX port map (\n\t\t") in
										match inst_id with
										"none"	-> raise (Failure ("Function instantiation id required: Lat"))
										| _		-> "\n\t"^inst_id^fcall_name ^
											String.concat ",\n\t\t" (List.map2 (fun x y -> x^y) mux_port (List.map expr actuals))^"\n\t)"
									)
					| _	-> raise (Failure ("undefined built-in function"))
					)
			
	in
	(* translate a statement *)
	let rec stmt = function
			  Block(s)		-> String.concat "" (List.map stmt s)
			| Expr(e)		-> if (eval_expr e < 0) then 
								raise (Failure ("Type mismatch in expression: " ^ Ast.string_of_expr e))
								else "\t" ^ expr e ^ ";\n"
			| If (pid, e, s, Block([]))	-> (match pid with
											 "none"	-> "\tif "^(expr e)^" then\n"^stmt s^"\tend if;\n"
											| _	-> "\n\tproc_"^pid^" : process (all)\n"^
												"\tbegin\n"^"\tif "^(expr e)^" then\n"^stmt s^
												"\tend if;\n\tend process;\n\n")
									   
			| If (pid, e, s1, s2)		-> (match pid with
											"none"	-> "\tif "^(expr e)^" then\n"^stmt s1^
																"\telse\n"^stmt s2^"\tend if;\n"
											| _	-> "\n\tproc_"^pid^" : process (all)\n"^
												"\tbegin\n"^"\tif "^(expr e)^" then\n"^stmt s1^
												"\telse\n"^stmt s2^"\tend if;\n\tend process;\n\n")
			| StError(e)	-> raise (Failure ("Syntax Error - semi-colon(;) is expected : " ^Ast.string_of_expr e))
			
	in
	(* translate a whole function *)
	(
		match out_opt with
			(* translate user-defined function declarations to component declarations in package file *)
			"pkg_decl"	-> (
					match (Ast.string_of_funtype func_decl.func_type) with
					"User "	-> 
						"component " ^ func_decl.func_name ^ " is\n" ^ "port (\n" ^ String.concat ";\n" 
						(List.map string_of_formal func_decl.formals) ^	"\n);\nend component " ^ 
						func_decl.func_name ^ ";\n" 
					| _	-> ""
					)
			(* translate user-defined functions into package body - ENTITY and ARCHITECTURE *)
			| "pkg_body"	-> (
					match (Ast.string_of_funtype func_decl.func_type) with
					"User "	-> 
						vhd_header_common ^ vhd_header_opt ^ "\n\n" ^
						"ENTITY " ^ func_decl.func_name ^ " is\n" ^ "port (\n" ^ String.concat ";\n" 
						(List.map string_of_formal func_decl.formals) ^	"\n);\nend " ^ func_decl.func_name ^ 
						";\n" ^ "\narchitecture RTL of " ^ func_decl.func_name ^ " is\n" ^	String.concat "" 
						(List.map string_of_variable func_decl.locals) ^ "begin\n" ^ 
						String.concat "" (List.map stmt func_decl.body) ^ "end RTL;\n" ^
						"\nconfiguration CFG_" ^ func_decl.func_name ^ " of " ^ func_decl.func_name ^ " is\n" ^
						"\tfor RTL\n" ^ "\tend for;\n" ^ "end " ^ "CFG_" ^ func_decl.func_name ^ ";\n"
					| _	-> ""
					)
			(* translate main functions into main VHDL *)
			| "main"	-> (
					match (Ast.string_of_funtype func_decl.func_type) with
					"Main "	-> 
						"ENTITY " ^ func_decl.func_name ^ "_top is\n" ^ "port (\n" ^ String.concat ";\n" 
						(List.map string_of_formal func_decl.formals) ^	"\n);\nend " ^ func_decl.func_name ^ 
						"_top;\n" ^ "\narchitecture RTL of " ^ func_decl.func_name ^ "_top is\n" ^	String.concat "" 
						(List.map string_of_variable func_decl.locals) ^ "begin\n" ^ 
						String.concat "" (List.map stmt func_decl.body) ^ "end RTL;\n"
					| _	-> ""
					)
			| "tbench"	-> (
					match (Ast.string_of_funtype func_decl.func_type) with
					"Main "	-> 
						"ENTITY " ^ func_decl.func_name ^ "_top_tb is\n\n" ^ "end ENTITY " ^ func_decl.func_name ^ "_top_tb;\n\n" ^
						"ARCHITECTURE behavioral of " ^ func_decl.func_name ^ "_top_tb is\n\n" ^
						"COMPONENT " ^ func_decl.func_name ^ "_top is\n" ^ "port (\n" ^ String.concat ";\n" 
						(List.map string_of_formal func_decl.formals) ^	"\n);\nend COMPONENT;\n\n" ^
						String.concat ";\n" (List.map string_of_formal_tb func_decl.formals) ^ ";\n\n" ^
						"begin\n\n" ^ "inst_" ^ func_decl.func_name ^ "_top : " ^ func_decl.func_name ^ "_top\n" ^ "\tport map(\n" ^
						"\t\t" ^ (String.concat ",\n\t\t" (List.map (fun x -> x.formal_name^" => "^x.formal_name) func_decl.formals)) ^ 
						"\n\t);\n" ^ "\n\n\nend;"
					| _	-> ""
					)
			| _	-> ""
		)
	
	in let env = { 
         function_decl = function_decls;
		 global_var = global_variables;
		 local_var = StringMap.empty } in
	
	(* Code executed to start the program *)
	let _ = try
		(StringMap.mem "main" function_decls)
	with Not_found -> raise (Failure ("no \"main\" function"))
    
	(* Compile the functions *)
	in 
		(
		match out_opt with
			(* translate global variables to variable declarations in package file *)
 			  "const"	-> file_header ^ vhd_header_common ^ "\npackage globals is\n" ^
							"\n-- Constants and global signals declaration\n" ^
							(String.concat "" (List.map string_of_variable (List.rev globals))) ^ "\n"
			| "pkg_decl" 	-> "-- Components declaration\n" ^(String.concat "\n" (List.map (translate env) (List.rev functions)))
								^ "end globals;\n"
			| "pkg_body" 	-> "\n\n-- Entities and architectures of the components\n" ^ 
							(String.concat "\n" (List.map (translate env) (List.rev functions))) ^ "\n"
			| "main" 	-> file_header ^ vhd_header_common ^ vhd_header_opt ^ "\n\n" ^
							(String.concat "\n" (List.map (translate env) (List.rev functions))) ^ "\n"
			| "tbench"	-> file_header ^ vhd_header_common ^ vhd_header_opt ^ "\n\n" ^
							(String.concat "\n" (List.map (translate env) (List.rev functions))) ^ "\n"
			| _			-> raise (Failure ("Input file is empty"))
		)














	
			
	
